print("import: 'CoolProp'")
import CoolProp

print("import: 'CoolProp.tests'")
import CoolProp.tests

print("import: 'CoolProp.Plots'")
import CoolProp.Plots

print("import: 'CoolProp.GUI'")
import CoolProp.GUI

